#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include "isoadc12.h"

/* The name this program was run with.  */
static char *program_name;

#define DEVICE "/dev/mem"
#define A460_PC104_IO_8_BASE (0xb2000000)
#define A460_PC104_IO_8_SIZE (0x10000)

#define CHANNEL_OPTION_DELIMITER (',')

#define LED1 (ISOADC12_LEDCTRL_LED1)
#define LED2 (ISOADC12_LEDCTRL_LED2)

static void *isoadc12_base;
#define isoadc12_readb(addr) ({*(volatile uint8_t *)((uintptr_t)isoadc12_base + addr);})
#define isoadc12_writeb(addr, data) ({ \
			*(volatile uint8_t *)((uintptr_t)isoadc12_base + addr) = data;})
#define isoadc12_readw(addr) ({*(volatile uint16_t *)((uintptr_t)isoadc12_base + addr);})
#define isoadc12_writew(addr, data) ({ \
			*(volatile uint16_t *)((uintptr_t)isoadc12_base + addr) = data;})

static int print_ch[ISOADC12_NR_CH_MAX];

static struct option const long_options[] =
{
  {"channel", required_argument, NULL, 'c'},
  {"range-10v", no_argument, NULL, 'r'},
  {"bipolar-negative", no_argument, NULL, 'b'},
  {"board-id", no_argument, NULL, 'B'},
  {"led1", required_argument, NULL, '1'},
  {"led2", required_argument, NULL, '2'},
  {"input", no_argument, NULL, 'i'},
  {"output", required_argument, NULL, 'o'},
  {"help", no_argument, NULL, 'h'},
  {NULL, 0, NULL, 0}
};

static void usage(int status)
{
	FILE *s;

	if (status == EXIT_SUCCESS)
		s = stdout;
	else
		s = stderr;

	fprintf(s, "Usage: %s [OPTION]...\n", program_name);
	fputs("\
  -c, --channel=CHANNEL      specified channel\n", s);
	fputs("\
  -r, --range-10v            set voltage range to 10V\n", s);
	fputs("\
  -b, --bipolar-negative     convert negative voltage\n", s);
	fputs("\
  -B, --board-id             display board id\n", s);
	fputs("\
  -1, --led1=STAT            control LED1\n\
                             0: LED_OFF, !0: LED_ON\n", s);
	fputs("\
  -2, --led2=STAT            control LED2\n\
                             0: LED_OFF, !0: LED_ON\n", s);
	fputs("\
  -i, --input                show isolate input data\n", s);
	fputs("\
  -o, --output=DATA          set isolate output data\n", s);
	fputs ("\
  -h, --help                 display this help and exit\n", s);

	exit(status);
}

static inline uint8_t _get_status(void)
{
	volatile uint8_t stat;

	/* we must read the register twice */
	stat = isoadc12_readb(ISOADC12_STAT);
	stat = isoadc12_readb(ISOADC12_STAT);

	return stat;
}

static inline uint8_t get_led_status(void)
{
	return (_get_status() & (ISOADC12_STAT_LEDS2 | ISOADC12_STAT_LEDS1));
}

static inline uint8_t get_board_id(void)
{
	uint8_t led = get_led_status();

	isoadc12_writeb(ISOADC12_LEDCTRL, ISOADC12_LEDCTRL_RDBID | led);

	return isoadc12_readb(ISOADC12_BID);
}

static inline int have_ext_channel(void)
{
	return !!(_get_status() & ISOADC12_STAT_OPT);
}

static inline uint8_t get_nr_channel(void)
{
	return (have_ext_channel() ? ISOADC12_NR_CH_ADC12_16 : ISOADC12_NR_CH_ADC12);
}

static void change_led_status(uint8_t leds, int status)
{
	uint8_t led;

	led = get_led_status();
	if (status)
		led |= leds;
	else
		led &= ~leds;

	isoadc12_writeb(ISOADC12_LEDCTRL, led);
}

static short read_ad(uint8_t ch, uint8_t scale)
{
	int retry;

	retry = 0;
	while (_get_status() & ISOADC12_STAT_BUSY) {
		usleep(ISOADC12_CONVERT_TIME);
		if (0 < retry++) {
			fputs("BUSY flag is not cleared.", stderr);
			exit(EXIT_FAILURE);
		}
	}

	isoadc12_writeb(ISOADC12_ADDCMD, scale | (ch & ISOADC12_ADDCMD_CH_MASK));
	retry = 0;
	while (!(_get_status() & ISOADC12_STAT_EOC)) {
		usleep(ISOADC12_CONVERT_TIME);
		if (0 < retry++) {
			fputs("EOC flag is not cleared.", stderr);
			exit(EXIT_FAILURE);
		}
	}

	return isoadc12_readw(ISOADC12_ADDR);
}

static double ad2voltage(int16_t ad, uint8_t scale)
{
	int rng = scale & ISOADC12_ADDCMD_RNG;
	int bip = scale & ISOADC12_ADDCMD_BIP;
	int vmax = rng ? ISOADC12_RNG1_VMAX : ISOADC12_RNG0_VMAX;
	int amax = bip ? ISOADC12_BIP1_AMAX : ISOADC12_BIP0_AMAX;
	double v;

	v = (double)ad * vmax / amax;

	return v;
}

static void print_ad(uint8_t scale)
{
	uint8_t i;
	int16_t ad;

	for (i = 0; i < ISOADC12_NR_CH_MAX; i++) {
		if (print_ch[i]) {
			ad = read_ad(i, scale);
			printf("ch%02d: %f\n", i, ad2voltage(ad, scale));
		}
	}
}

static inline uint8_t read_isolate_port(void)
{
	return isoadc12_readb(ISOADC12_PI);
}

static inline void write_isolate_port(uint8_t data)
{
	isoadc12_writeb(ISOADC12_PO, data);
}

static int xmmap(void)
{
	int fd;

	fd = open(DEVICE, O_RDWR | O_SYNC);
	if (fd < 0) {
		perror("open");
		return -1;
	}

	isoadc12_base = mmap(0x0, A460_PC104_IO_8_SIZE,PROT_READ |
			     PROT_WRITE, MAP_SHARED, fd, A460_PC104_IO_8_BASE);
	if (isoadc12_base == MAP_FAILED) {
		perror("mmap");
		close(fd);
		return -1;
	}

	close(fd);

	return 0;
}

static void xmunmap(void)
{
	munmap(isoadc12_base, A460_PC104_IO_8_SIZE);
}

static int xstrtoul(const char *str, unsigned long *val)
{
	char *endptr;

	if (*str == '\0')
		return -1;

	errno = 0;
	*val = strtoul(str, &endptr, 0);
	if (errno || *endptr)
		return -1;

	return 0;
}

static int parse_channel_options(char *options)
{
	uint8_t nr_ch;
	int i;

	for (i = 0; i < ISOADC12_NR_CH_MAX; i++)
		print_ch[i] = 0;

	nr_ch = get_nr_channel();
	while (options) {
		char *delim;
		unsigned long ch;
		int ret;

		delim = strchr(options, CHANNEL_OPTION_DELIMITER);
		if (delim)
			*delim = '\0';

		ret = xstrtoul(options, &ch);
		if ((ret < 0) || (nr_ch <= ch)) {
			fprintf(stderr, "Channel %s not supported.\n", options);
			return -1;
		}
		print_ch[ch] = 1;

		if (!delim)
			break;

		options = ++delim;
	}

	return 0;
}

int main(int argc, char *argv[])
{
	int ret;
	uint8_t scale = 0;
	unsigned long opt;

	ret = xmmap();
	if (ret)
		return EXIT_FAILURE;
	ret = atexit(xmunmap);
	if (ret)
		return EXIT_FAILURE;

	while (1) {
		int c = getopt_long (argc, argv, "c:rbB1:2:io:h", long_options, NULL);
		if (c == -1)
			break;

		switch (c) {
		case 'c':
			ret = parse_channel_options(optarg);
			if (ret < 0)
				exit(EXIT_FAILURE);
			break;
		case 'r':
			scale |= ISOADC12_ADDCMD_RNG;
			break;
		case 'b':
			scale |= ISOADC12_ADDCMD_BIP;
			break;
		case '1':
			ret = xstrtoul(optarg, &opt);
			if (ret < 0) {
				fprintf(stderr, "invalid led status: %s\n", optarg);
				return EXIT_FAILURE;
			}
			change_led_status(LED1, !!opt);
			break;
		case '2':
			ret = xstrtoul(optarg, &opt);
			if (ret < 0) {
				fprintf(stderr, "invalid led status: %s\n", optarg);
				return EXIT_FAILURE;
			}
			change_led_status(LED2, !!opt);
			break;
		case 'B':
			printf("Board ID: 0x%x\n", get_board_id());
			break;
		case 'i':
			printf("Data: 0x%x\n", read_isolate_port());
			break;
		case 'o':
			ret = xstrtoul(optarg, &opt);
			if (ret < 0) {
				fprintf(stderr, "invalid output data: %s\n", optarg);
				return EXIT_FAILURE;
			}
			if (ISOADC12_PO_MAX < opt) {
				fprintf(stderr, "out of range(data=%ld)\n", opt);
				return EXIT_FAILURE;
			}
			write_isolate_port(opt);
			break;
		case 'h':
			usage(EXIT_SUCCESS);
			break;
		case '?':
		default:
			usage(EXIT_FAILURE);
		}
	}

	print_ad(scale);

	return EXIT_SUCCESS;
}
